/*
*   $Id: readtags.c,v 1.24 2003/07/20 18:46:30 darren Exp $
*
*   Copyright (c) 1996-2003, Darren Hiebert
*
*   This source code is released into the public domain.
*
*   This module contains functions for reading tag files.
*
*   ### I'm not sure how to license this (because I used some of Darren's code
*   ### but almost everything is a rewrite and mostly I just looked at his code
*   ### Some advice needed, Jesse van den Kieboom
*/

/*
*   INCLUDE FILES
*/
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <errno.h>
#include <sys/types.h>	/* to declare off_t */
#include <gedit/gedit-debug.h>
#include "gedit-code-listing-plugin-parser.h"

#define GEDIT_CODE_LISTING_PLUGIN_PARSER_GET_PRIVATE(object)(G_TYPE_INSTANCE_GET_PRIVATE ((object), GEDIT_TYPE_CODE_LISTING_PLUGIN_PARSER, GeditCodeListingPluginParserPrivate))

static GObjectClass *parent_class = NULL;
static GType type;

/* Private functions */
gboolean gedit_code_listing_plugin_parser_parse_tag(GeditCodeListingPluginParser *parser, tagEntry **entry);
void gedit_code_listing_plugin_parser_free_tag(tagEntry *entry);
void gedit_code_listing_plugin_parser_free_tags(GList *tags);
void gedit_code_listing_plugin_parser_parse_tag_ext(tagEntry *entry, gchar *ext);
void gedit_code_listing_plugin_parser_parse_header(GeditCodeListingPluginParser *parser);
gboolean gedit_code_listing_plugin_parser_idle_parse(GeditCodeListingPluginParser *parser);
gboolean gedit_code_listing_plugin_parser_idle_search(GeditCodeListingPluginParser *parser);

/* Signals */
enum {
	TAG_PARSED = 0,
	PARSE_DONE,
	TAG_FOUND,
	SEARCH_DONE,
	LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = { 0 };

struct _GeditCodeListingPluginParserPrivate
{
	GeditCodeListingPluginParserTagFile tagfile;
	
	gint idle_parse;
	gint idle_search;
	GList *current_tag;
	gchar *search_for;
	long search_position;
};

static void
gedit_code_listing_plugin_parser_finalize (GObject *object)
{
  GeditCodeListingPluginParser *parser = GEDIT_CODE_LISTING_PLUGIN_PARSER(object);
  GeditCodeListingPluginParserTagFile *tagFile = &(parser->priv->tagfile);
  
  if (parser->priv->idle_parse != -1) {
    g_source_remove(parser->priv->idle_parse);
    parser->priv->idle_parse = -1;
  }
  if (parser->priv->idle_search != -1) {
    g_source_remove(parser->priv->idle_search);
    parser->priv->idle_search = -1;
  }
  
  g_free(tagFile->path);
  g_free(tagFile->program.author);
  g_free(tagFile->program.name);
  g_free(tagFile->program.url);
  g_free(tagFile->program.version);
  
  if (tagFile->fd != NULL) {
    fclose(tagFile->fd);
  }
  
  gedit_code_listing_plugin_parser_free_tags(tagFile->tags);    
    
	G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void 
gedit_code_listing_plugin_parser_class_init (GeditCodeListingPluginParserClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

	parent_class = g_type_class_peek_parent (klass);
	
	object_class->finalize = gedit_code_listing_plugin_parser_finalize;
	//object_class->get_property = gedit_code_listing_plugin_panel_get_property;
	//object_class->set_property = gedit_code_listing_plugin_panel_set_property;
	
	/*g_object_class_install_property (object_class,
					 PROP_WINDOW,
					 g_param_spec_object ("window",
							 "Window",
							 "The GeditWindow this GeditCodeListingPluginPanel is associated with",
							 GEDIT_TYPE_WINDOW,
							 G_PARAM_READWRITE |
							 G_PARAM_CONSTRUCT_ONLY));*/

	signals[TAG_PARSED] =
	    g_signal_new ("tag-parsed",
			  G_OBJECT_CLASS_TYPE (object_class),
			  G_SIGNAL_RUN_LAST,
			  G_STRUCT_OFFSET (GeditCodeListingPluginParserClass, tag_parsed),
			  NULL, NULL,
			  g_cclosure_marshal_VOID__POINTER,
			  G_TYPE_NONE, 
			  1, 
			  G_TYPE_POINTER);
			  
	signals[PARSE_DONE] =
	    g_signal_new ("parse-done",
			  G_OBJECT_CLASS_TYPE (object_class),
			  G_SIGNAL_RUN_LAST,
			  G_STRUCT_OFFSET (GeditCodeListingPluginParserClass, parse_done),
			  NULL, NULL,
			  g_cclosure_marshal_VOID__VOID,
			  G_TYPE_NONE, 
			  0);
			  
	signals[TAG_FOUND] =
	    g_signal_new ("tag-found",
			  G_OBJECT_CLASS_TYPE (object_class),
			  G_SIGNAL_RUN_LAST,
			  G_STRUCT_OFFSET (GeditCodeListingPluginParserClass, tag_found),
			  NULL, NULL,
			  g_cclosure_marshal_VOID__POINTER,
			  G_TYPE_NONE, 
			  1, 
			  G_TYPE_POINTER);
			  
	signals[SEARCH_DONE] =
	    g_signal_new ("search-done",
			  G_OBJECT_CLASS_TYPE (object_class),
			  G_SIGNAL_RUN_LAST,
			  G_STRUCT_OFFSET (GeditCodeListingPluginParserClass, search_done),
			  NULL, NULL,
			  g_cclosure_marshal_VOID__VOID,
			  G_TYPE_NONE, 
			  0);
			  
	g_type_class_add_private (object_class, sizeof(GeditCodeListingPluginParserPrivate));
}

static void
gedit_code_listing_plugin_parser_init (GeditCodeListingPluginParser *parser)
{	
  GeditCodeListingPluginParserTagFile *tagFile;
  
	gedit_debug (DEBUG_PLUGINS);
	parser->priv = GEDIT_CODE_LISTING_PLUGIN_PARSER_GET_PRIVATE (parser);
	tagFile = &(parser->priv->tagfile);
	
	tagFile->path = NULL;
	tagFile->fd = NULL;
	tagFile->tags = NULL;
	tagFile->ntags = 0;
  tagFile->format = 1;
  tagFile->sortMethod = TAG_UNSORTED;
  tagFile->program.author = NULL;
  tagFile->program.name = NULL;
  tagFile->program.url = NULL;
  tagFile->program.version = NULL;
  
  parser->priv->idle_parse = -1;
  parser->priv->idle_search = -1;
  parser->priv->search_for = NULL;
}

GType
gedit_code_listing_plugin_parser_get_type (void)
{
	return type;
}

GType
gedit_code_listing_plugin_parser_register_type (GTypeModule *module)
{
	static const GTypeInfo our_info =
	{
		sizeof (GeditCodeListingPluginParserClass),
		NULL, /* base_init */
		NULL, /* base_finalize */
		(GClassInitFunc) gedit_code_listing_plugin_parser_class_init,
		NULL,
		NULL, /* class_data */
		sizeof (GeditCodeListingPluginParser),
		0, /* n_preallocs */
		(GInstanceInitFunc) gedit_code_listing_plugin_parser_init
	};

	gedit_debug_message (DEBUG_PLUGINS, "Registering GeditCodeListingPluginParser");

	type = g_type_module_register_type (module,
					    G_TYPE_OBJECT,
					    "GeditCodeListingPluginParser",
					    &our_info, 
					    0);
	return type;
}

GeditCodeListingPluginParser *
gedit_code_listing_plugin_parser_new(gchar *path)
{
  GeditCodeListingPluginParser *result;

  result = GEDIT_CODE_LISTING_PLUGIN_PARSER(g_object_new(GEDIT_TYPE_CODE_LISTING_PLUGIN_PARSER, NULL));
  
  if (path != NULL) {
    result->priv->tagfile.path = g_strdup(path);
  }
  
  return result;
}


/* Private functions */
const char *const EMPTY_STRING = "";
const char *const PSEUDO_TAG_PREFIX = "!_";

void
gedit_code_listing_plugin_parser_free_tag(tagEntry *entry) {
  g_free(entry->name);
  g_free(entry->file);
  g_free(entry->address.pattern);
  g_free(entry->kind);
  g_hash_table_destroy(entry->fields);
}

void 
gedit_code_listing_plugin_parser_free_tags(GList *tags) {
  GList *item = tags;
  
  while (item) {
    gedit_code_listing_plugin_parser_free_tag((tagEntry *)(item->data));
    item = item->next;
  }
  
  g_list_free(tags);
}

void
gedit_code_listing_plugin_parser_parse_tag_ext(tagEntry *entry, gchar *ext) {
  gchar **extFields, *colon, *key, *value;
  int i;
  
  extFields = g_strsplit(ext, "\t", 0);
  entry->kind = NULL;
  
  for (i = 0; extFields[i]; i++) {
    colon = strchr(extFields[i], ':');
    
    if (colon == NULL) {
      g_free(entry->kind);
      entry->kind = g_strdup(extFields[i]);
    } else {
      *colon = '\0';
      
      key = extFields[i];
      value = colon + 1;
      
      if (strcmp(key, "kind") == 0) {
        g_free(entry->kind);
        entry->kind = g_strdup(value);
      } else if (strcmp(key, "file") == 0) {
        entry->fileScope = 1;
      } else if (strcmp(key, "line") == 0) {
        entry->address.lineNumber = atol(value);
      } else {
        g_hash_table_insert(entry->fields, g_strdup(key), g_strdup(value));
      }
    }
  }
  
  g_strfreev(extFields);
}

gboolean
gedit_code_listing_plugin_parser_parse_tag(GeditCodeListingPluginParser *parser, tagEntry **entry) {
  GeditCodeListingPluginParserTagFile *tagFile;
  char buf[256];
  gchar **tabsep, *pat, delim;
  GString *str;
    
  tagFile = &(parser->priv->tagfile);
  buf[0] = '\0';
  
  if (tagFile->fd == NULL || feof(tagFile->fd)) {
    return FALSE;
  }
  
  str = g_string_new(NULL);
  
  /* Read one line from the file */
  while (!feof(tagFile->fd) && buf[strlen(buf) - 1] != '\n') {
    fgets((char *)(&buf), 255, tagFile->fd);
    str = g_string_append(str, (gchar *)(&buf));
  }
    
  if (str->str[str->len - 1] == '\n') {
    g_string_set_size(str, str->len - 1);
  }
  
  /* Now that we have our string */
  tabsep = g_strsplit(str->str, "\t", 3);
  
  if (tabsep[0] == NULL || tabsep[1] == NULL || tabsep[2] == NULL) {
    /* Well it should at least have 3 items */
    (*entry) = NULL;
  } else {
    (*entry) = g_new0(tagEntry, 1);
    (*entry)->fields = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_free);

    (*entry)->name = g_strdup(tabsep[0]);
    (*entry)->file = g_strdup(tabsep[1]);

    pat = tabsep[2];
    
    if (tabsep[2][0] == '/' || tabsep[2][0] == '?') {
      /* Okay, so it's a pattern */
      delim = tabsep[2][0];
      pat = tabsep[2];
      
      do {
        pat = strchr(pat + 1, delim);
      } while (pat != NULL && *(pat - 1) == '\\');
      
      if (pat != NULL) {
        ++pat;
        (*entry)->address.lineNumber = 0;
        (*entry)->address.pattern = g_strndup(tabsep[2], pat - tabsep[2]);
      } else {
        /* Invalid field, does not end */
      }
    } else if (isdigit(tabsep[2][0])) {
      (*entry)->address.lineNumber = 0;
      
      while (*pat != '\0' && isdigit(*pat)) {
        (*entry)->address.lineNumber = ((*entry)->address.lineNumber * 10) + g_ascii_digit_value(*pat);
        ++pat;
      }
      
      (*entry)->address.pattern = g_strdup(tabsep[2]);
    } else {
      /* Invalid field */
    }
    
    /* pat now points to the end of the third field */
    if (g_str_has_prefix(pat, ";\"")) {
      /* Parse extension fields */
      gedit_code_listing_plugin_parser_parse_tag_ext((*entry), pat + 2);
    }
  }

  g_strfreev(tabsep);
  g_string_free(str, TRUE);
  
  return TRUE;
}

void 
gedit_code_listing_plugin_parser_parse_header(GeditCodeListingPluginParser *parser) {
  GeditCodeListingPluginParserTagFile *tagFile;
  gchar *key, *value;
  tagEntry *entry;
  
  tagFile = &(parser->priv->tagfile);
  
  while (1) {
	  gedit_code_listing_plugin_parser_parse_tag(parser, &entry);
	  
  	if (entry == NULL || !g_str_has_prefix(entry->name, PSEUDO_TAG_PREFIX)) {
	    break;
	  } else {
	    key = entry->name + strlen(PSEUDO_TAG_PREFIX);
	    value = entry->file;
	    
	    if (strcmp(key, "TAG_FILE_SORTED") == 0) {
 		    tagFile->sortMethod = (sortType) atoi (value);
 		  } else if (strcmp (key, "TAG_FILE_FORMAT") == 0) {
		    tagFile->format = atoi (value);
	    } else if (strcmp (key, "TAG_PROGRAM_AUTHOR") == 0) {
		    tagFile->program.author = g_strdup(value);
	    } else if (strcmp (key, "TAG_PROGRAM_NAME") == 0) {
		    tagFile->program.name = g_strdup(value);
	    } else if (strcmp (key, "TAG_PROGRAM_URL") == 0) {
		    tagFile->program.url = g_strdup(value);
	    } else if (strcmp (key, "TAG_PROGRAM_VERSION") == 0) {
		    tagFile->program.version = g_strdup(value);
		  }
    }
    
    gedit_code_listing_plugin_parser_free_tag(entry);
  }
}

gboolean
gedit_code_listing_plugin_parser_idle_search(GeditCodeListingPluginParser *parser) {
  gchar *low1;
  tagEntry *entry;
  int counter;
  gboolean found = FALSE;
  
  counter = 0;
  
  if (parser->priv->current_tag == NULL) {
    g_signal_emit(G_OBJECT(parser), signals[SEARCH_DONE], 0);
    g_free(parser->priv->search_for);
    parser->priv->idle_search = -1;
    return FALSE;
  }

  while (!found && parser->priv->current_tag != NULL && counter < 50) {
    entry = (tagEntry *)(parser->priv->current_tag->data);
    
    if (parser->priv->search_position < parser->priv->tagfile.ntags) {
      parser->priv->search_position = parser->priv->search_position + 1;
    }
    
    low1 = g_ascii_strdown(entry->name, -1);
    parser->priv->current_tag = parser->priv->current_tag->next;
    
    found = (g_strstr_len(low1, strlen(low1), parser->priv->search_for) != NULL);
	  g_free(low1);
	  counter++;
	}

  if (!found) {	
    g_signal_emit(G_OBJECT(parser), signals[TAG_FOUND], 0, NULL);
  } else {
    g_signal_emit(G_OBJECT(parser), signals[TAG_FOUND], 0, entry);
  }
  
  return TRUE;
}

gboolean
gedit_code_listing_plugin_parser_idle_parse(GeditCodeListingPluginParser *parser) {
  tagEntry *entry;
  GeditCodeListingPluginParserTagFile *tagFile = &(parser->priv->tagfile);
  
  if (gedit_code_listing_plugin_parser_parse_tag(parser, &entry) != FALSE) {
    if (entry) {
      tagFile->tags = g_list_prepend(tagFile->tags, entry);
      tagFile->position = ftell(tagFile->fd);
      g_signal_emit(G_OBJECT(parser), signals[TAG_PARSED], 0, entry);
    }
    return TRUE;
  } else {
    parser->priv->idle_parse = -1;
    fclose(tagFile->fd);
    tagFile->fd = NULL;
    g_signal_emit(G_OBJECT(parser), signals[PARSE_DONE], 0);
    tagFile->ntags = g_list_length(tagFile->tags);
    return FALSE;
  }
}

/* Public functions */
gboolean 
gedit_code_listing_plugin_parser_parse(GeditCodeListingPluginParser *parser) {
  GeditCodeListingPluginParserTagFile *tagFile = &(parser->priv->tagfile);
  
  /* Check whether file is already loaded or no path is specified */
  if (tagFile->tags != NULL || tagFile->path == NULL) {
    return FALSE;
  }
  
  /* Open the file */
	tagFile->fd = fopen(tagFile->path, "r");
	
	if (tagFile->fd == NULL) {
	  return FALSE;
	} else {
	  /* Get file size */
	  fseek(tagFile->fd, 0, SEEK_END);
	  tagFile->size = ftell(tagFile->fd);
	  rewind(tagFile->fd);
	  
	  /* Process header information */
	  gedit_code_listing_plugin_parser_parse_header(parser);

    parser->priv->idle_parse = g_idle_add((GSourceFunc)gedit_code_listing_plugin_parser_idle_parse, parser);
	}
	
	return TRUE;
}

void
gedit_code_listing_plugin_parser_search(GeditCodeListingPluginParser *parser, gchar *key) {
  if (parser->priv->idle_parse == -1) {
	  parser->priv->current_tag = gedit_code_listing_plugin_parser_tags(parser);
	  parser->priv->search_for = g_ascii_strdown(key, -1);
	  parser->priv->search_position = 0;
	  parser->priv->idle_search = g_idle_add((GSourceFunc)gedit_code_listing_plugin_parser_idle_search, parser);
	}
}

void
gedit_code_listing_plugin_parser_cancel(GeditCodeListingPluginParser *parser) {
  GeditCodeListingPluginParserTagFile *tagFile = &(parser->priv->tagfile);
  
  if (parser->priv->idle_parse != -1) {
    g_source_remove(parser->priv->idle_parse);
    parser->priv->idle_parse = -1;
    fclose(tagFile->fd);
    tagFile->fd = NULL;
    g_signal_emit(G_OBJECT(parser), signals[PARSE_DONE], 0);
    tagFile->ntags = g_list_length(tagFile->tags);
  } else if (parser->priv->idle_search != -1) {
    g_source_remove(parser->priv->idle_search);
    g_signal_emit(G_OBJECT(parser), signals[SEARCH_DONE], 0);
    g_free(parser->priv->search_for);
    parser->priv->idle_search = -1;
  }
}

GList *
gedit_code_listing_plugin_parser_tags(GeditCodeListingPluginParser *parser) {
  return parser->priv->tagfile.tags;
}

void
gedit_code_listing_plugin_parser_parse_progress(GeditCodeListingPluginParser *parser, long *position, long *size) {
  GeditCodeListingPluginParserTagFile *tagFile = &(parser->priv->tagfile);
 
  if (tagFile->fd != NULL) { 
    *position = tagFile->position;
    *size = tagFile->size;
  } else {
    *position = 0;
    *size = 0;
  }
}

void
gedit_code_listing_plugin_parser_search_progress(GeditCodeListingPluginParser *parser, long *position, long *size) {
  if (parser->priv->idle_search != -1) {
    *position = parser->priv->search_position;
    *size = parser->priv->tagfile.ntags;
  } else {
    *position = 0;
    *size = 0;
  }
}

gchar *
gedit_code_listing_plugin_parser_dirname(GeditCodeListingPluginParser *parser) {
  return g_path_get_dirname(parser->priv->tagfile.path);
}
