/*
 * gedit-code-listing-plugin.c
 * 
 * Copyright (C) 2005 - Jesse van den Kieboom
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "gedit-code-listing-plugin.h"
#include "gedit-code-listing-plugin-panel.h"
#include "gedit-code-listing-plugin-parser.h"

#include <glib/gi18n-lib.h>
#include <gmodule.h>

#include <gedit/gedit-debug.h>

#define WINDOW_DATA_KEY "GeditCodeListingPluginWindowData"

#define GEDIT_CODE_LISTING_PLUGIN_GET_PRIVATE(object)(G_TYPE_INSTANCE_GET_PRIVATE ((object), GEDIT_TYPE_CODE_LISTING_PLUGIN, GeditCodeListingPluginPrivate))

struct _GeditCodeListingPluginPrivate
{
	gpointer dummy;
};

static GType gedit_code_listing_plugin_type = 0;

GType
gedit_code_listing_plugin_get_type (void)
{
	return gedit_code_listing_plugin_type;
}

static void     gedit_code_listing_plugin_init              (GeditCodeListingPlugin        *self);
static void     gedit_code_listing_plugin_class_init        (GeditCodeListingPluginClass *klass);
static gpointer gedit_code_listing_plugin_parent_class = NULL;
static void     gedit_code_listing_plugin_class_intern_init (gpointer klass)
{
	gedit_code_listing_plugin_parent_class = g_type_class_peek_parent (klass);
	gedit_code_listing_plugin_class_init ((GeditCodeListingPluginClass *) klass);
}

G_MODULE_EXPORT GType
register_gedit_plugin (GTypeModule *module)
{
	static const GTypeInfo our_info =
	{
		sizeof (GeditCodeListingPluginClass),
		NULL, /* base_init */
		NULL, /* base_finalize */
		(GClassInitFunc) gedit_code_listing_plugin_class_intern_init,
		NULL,
		NULL, /* class_data */
		sizeof (GeditCodeListingPlugin),
		0, /* n_preallocs */
		(GInstanceInitFunc) gedit_code_listing_plugin_init
	};

	gedit_debug_message (DEBUG_PLUGINS, "Registering GeditCodeListingPlugin");

	/* Initialise the i18n stuff */
	bindtextdomain (GETTEXT_PACKAGE, GEDIT_LOCALEDIR);
	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");

	gedit_code_listing_plugin_type = g_type_module_register_type (module,
					    GEDIT_TYPE_PLUGIN,
					    "GeditCodeListingPlugin",
					    &our_info,
					    0);
	
	gedit_code_listing_plugin_panel_register_type(module);
	gedit_code_listing_plugin_parser_register_type(module);
	return gedit_code_listing_plugin_type;
}

static void
gedit_code_listing_plugin_init (GeditCodeListingPlugin *plugin)
{
	plugin->priv = GEDIT_CODE_LISTING_PLUGIN_GET_PRIVATE (plugin);

	gedit_debug_message (DEBUG_PLUGINS, "GeditCodeListingPlugin initializing");
}

static void
gedit_code_listing_plugin_finalize (GObject *object)
{
/*
	GeditCodeListingPlugin *plugin = GEDIT_CODE_LISTING_PLUGIN (object);
*/
	gedit_debug_message (DEBUG_PLUGINS, "GeditCodeListingPlugin finalizing");

	G_OBJECT_CLASS (gedit_code_listing_plugin_parent_class)->finalize (object);
}

static void
impl_activate (GeditPlugin *plugin,
	       GeditWindow *window)
{
	GeditPanel *side_panel;
	GtkWidget *code_listing_panel;
	
	gedit_debug (DEBUG_PLUGINS);
	
	g_return_if_fail (g_object_get_data (G_OBJECT (window), WINDOW_DATA_KEY) == NULL);
	
	side_panel = gedit_window_get_side_panel (window);
	code_listing_panel = gedit_code_listing_plugin_panel_new (window);
	
	gedit_panel_add_item_with_stock_icon (side_panel, 
			      code_listing_panel, 
			      "Code listing", 
			      GTK_STOCK_INDEX);
			      
	g_object_set_data (G_OBJECT (window), 
			   WINDOW_DATA_KEY,
			   code_listing_panel);
}

static void
impl_deactivate	(GeditPlugin *plugin,
		 GeditWindow *window)
{
	GeditPanel *side_panel;
	gpointer data;
	
	gedit_debug (DEBUG_PLUGINS);
	
	data = g_object_get_data (G_OBJECT (window), WINDOW_DATA_KEY);
	g_return_if_fail (data != NULL);
	
	side_panel = gedit_window_get_side_panel (window);

	gedit_panel_remove_item (side_panel, 
			      	 GTK_WIDGET (data));
			      
	g_object_set_data (G_OBJECT (window), 
			   WINDOW_DATA_KEY,
			   NULL);
}

static void
impl_update_ui	(GeditPlugin *plugin,
		 GeditWindow *window)
{
	gedit_debug (DEBUG_PLUGINS);
}

/*
static GtkWidget *
impl_create_configure_dialog (GeditPlugin *plugin)
{
	* Implements this function only and only if the plugin
	* is configurable. Otherwise you can safely remove it. *
}
*/

static void
gedit_code_listing_plugin_class_init (GeditCodeListingPluginClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);
	GeditPluginClass *plugin_class = GEDIT_PLUGIN_CLASS (klass);

	object_class->finalize = gedit_code_listing_plugin_finalize;

	plugin_class->activate = impl_activate;
	plugin_class->deactivate = impl_deactivate;
	plugin_class->update_ui = impl_update_ui;

	/* Only if the plugin is configurable */
	/* plugin_class->create_configure_dialog = impl_create_configure_dialog; */

	g_type_class_add_private (object_class, sizeof (GeditCodeListingPluginPrivate));
}
